// *********************************************************************
// BrianHG's Experimental conversion of sin_cos.vhd
// into System Verilog.
// April 20, 2022
// *********************************************************************

parameter N = 20;

module sin_cos_bg (
   input               clk    ,
   input               de_in  ,
   input       [N-1:0] angle  ,
   output wire         de_out ,
   output wire [N-1:0] sine   ,
   output wire [N-1:0] cosine  );

const integer angles [0:15] = '{77376, 40884, 20753, 10417,
                                 5213,  2607,  1304,   652,
                                  326,   163,    81,    41,
                                   20,    10,     5,     3};

reg                de_sr     [0:16] = '{default:0};
reg signed [N-1:0] remainder [0:16] = '{default:0};
reg signed [N-1:0] x         [0:16] = '{default:0};
reg signed [N-1:0] y         [0:16] = '{default:0};

assign de_out = de_sr[16] ;
assign sine   = x    [16] ;
assign cosine = y    [16] ;

always @(posedge clk) begin

        for (int i=0;i<=15;i=i+1) begin

               de_sr[i+1] <= de_sr[i];
               if (remainder[i] < 0) begin
                   remainder[i+1] <= remainder[i] + angles[i];
                   x[i+1]   <= x[i] + (y[i] >>> (i+1)); // >>> instead of >> to keep the original sign of the x/y[i].
                   y[i+1]   <= y[i] - (x[i] >>> (i+1));
               end else begin
                   remainder[i+1] <= remainder[i] - angles[i];
                   x[i+1]   <= x[i] - (y[i] >>> (i+1));
                   y[i+1]   <= y[i] + (x[i] >>> (i+1));
               end

        end // for loop

de_sr[0] <= de_in;

    if (de_in) begin
    
       case (angle[N-1:N-2])
         2'b00   : begin 
                         x[0] <=  (N)'( 300000);
                         y[0] <=  (N)'( 300000);
                         remainder[0] <= (N)'(angle - 'h20000) ;
                   end
         2'b01   : begin 
                         x[0] <=  (N)'(-300000);
                         y[0] <=  (N)'( 300000);
                         remainder[0] <= (N)'(angle - 'h60000) ;
                   end
         2'b10   : begin
                         x[0] <=  (N)'(-300000);
                         y[0] <=  (N)'(-300000);
                         remainder[0] <= (N)'(angle - 'hA0000) ;
                   end
         default : begin 
                         x[0] <=  (N)'( 300000);
                         y[0] <=  (N)'(-300000);
                         remainder[0] <= (N)'(angle - 'hE0000) ;
                   end
       endcase
    
    end // if de_in

end // always clk
endmodule
